# distutils: language = c++

from libc.stdint cimport uint32_t
from libcpp.memory cimport shared_ptr
from libcpp.map cimport map as cppmap
from libcpp.string cimport string
from cython.operator cimport dereference as deref
cimport decl

####################################
# Enums                            #
####################################

cpdef enum VariableType:
    Continuous = <uint32_t> decl.Continuous
    Integer = <uint32_t> decl.Integer
    Binary = <uint32_t> decl.Binary

cpdef enum Sense:
    Minimize = <uint32_t> decl.Minimize
    Maximize = <uint32_t> decl.Maximize

cpdef enum Relation:
    LessEqual = <uint32_t> decl.LessEqual
    Equal = <uint32_t> decl.Equal
    GreaterEqual = <uint32_t> decl.GreaterEqual

cpdef enum Preference:
    Any = <uint32_t> decl.Any
    Scip = <uint32_t> decl.Scip
    Gurobi = <uint32_t> decl.Gurobi
    Cplex = <uint32_t> decl.Cplex

####################################
# Classes                          #
####################################

cdef class Solution:

    cdef decl.Solution* p

    def __cinit__(self, size):
        self.p = new decl.Solution(size)

    def __dealloc__(self):
        del self.p

    def __len__(self):
        return self.p.size()

    def __getitem__(self, i):
        if i < 0 or i >= self.p.size():
            raise IndexError()
        return self.p[0][i]

    def __setitem__(self, i, value):
        if i < 0 or i >= self.p.size():
            raise IndexError()
        self.p[0][i] = value

    def resize(self, size):
        self.p.resize(size)

    def get_value(self):
        return self.p.getValue()

    def set_value(self, value):
        self.p.setValue(value)

cdef class QuadraticObjective:

    cdef decl.QuadraticObjective* p

    def __cinit__(self, size = 0):
        self.p = new decl.QuadraticObjective(size)

    def __dealloc__(self):
        del self.p

    def set_constant(self, value):
        self.p.setConstant(value)

    def get_constant(self):
        return self.p.getConstant()

    def set_coefficient(self, i, value):
        self.p.setCoefficient(i, value)

    def get_coefficients(self):
        return self.p.getCoefficients()

    def set_quadratic_coefficient(self, i, j, value):
        self.p.setQuadraticCoefficient(i, j, value)

    def set_sense(self, sense):
        self.p.setSense(sense)

    def get_sense(self):
        return Sense(self.p.getSense())

    def resize(self, size):
        self.p.resize(size)

    def __len__(self):
        return self.p.size()

cdef class LinearObjective:

    cdef decl.LinearObjective* p

    def __cinit__(self, size = 0):
        self.p = new decl.LinearObjective(size)

    def __dealloc__(self):
        del self.p

    def set_constant(self, value):
        self.p.setConstant(value)

    def get_constant(self):
        return self.p.getConstant()

    def set_coefficient(self, i, value):
        self.p.setCoefficient(i, value)

    def get_coefficients(self):
        return self.p.getCoefficients()

    def set_sense(self, sense):
        self.p.setSense(sense)

    def get_sense(self):
        return Sense(self.p.getSense())

    def resize(self, size):
        self.p.resize(size)

    def __len__(self):
        return self.p.size()

cdef class LinearConstraint:

    cdef decl.LinearConstraint* p

    def __cinit__(self):
        self.p = new decl.LinearConstraint()

    def __dealloc__(self):
        del self.p

    def set_coefficient(self, i, value):
        self.p.setCoefficient(i, value)

    def get_coefficients(self):
        return self.p.getCoefficients()

    def set_relation(self, relation):
        self.p.setRelation(relation)

    def set_value(self, value):
        self.p.setValue(value)

    def get_relation(self):
        return Relation(self.p.getRelation())

    def get_value(self):
        return self.p.getValue()

    def is_violated(self, Solution solution):
        return self.p.isViolated(solution.p[0])

cdef class LinearConstraints:

    cdef decl.LinearConstraints* p

    def __cinit__(self):
        self.p = new decl.LinearConstraints()

    def __dealloc__(self):
        del self.p

    def clear(self):
        self.p.clear()

    def add(self, LinearConstraint constraint):
        self.p.add(constraint.p[0])

    def add_all(self, LinearConstraints constraints):
        self.p.addAll(constraints.p[0])

    def __len__(self):
        return self.p.size()

cdef class LinearSolver:

    cdef shared_ptr[decl.LinearSolverBackend] p
    cdef unsigned int num_variables

    def __cinit__(
            self,
            num_variables,
            default_variable_type,
            dict variable_types=None,
            Preference preference=Preference.Any):
        cdef decl.SolverFactory factory
        cdef cppmap[unsigned int, decl.VariableType] vtypes
        if variable_types is not None:
            for k, v in variable_types.items():
                vtypes[k] = v
        self.p = factory.createLinearSolverBackend(preference)
        self.num_variables = num_variables
        deref(self.p).initialize(num_variables, default_variable_type, vtypes)

    def set_objective(self, LinearObjective objective):
        deref(self.p).setObjective(objective.p[0])

    def set_constraints(self, LinearConstraints constraints):
        deref(self.p).setConstraints(constraints.p[0])

    def add_constraint(self, LinearConstraint constraint):
        deref(self.p).addConstraint(constraint.p[0])

    def set_timeout(self, timeout):
        deref(self.p).setTimeout(timeout)

    def set_optimality_gap(self, gap, absolute=False):
        deref(self.p).setOptimalityGap(gap, absolute)

    def set_num_threads(self, num_threads):
        deref(self.p).setNumThreads(num_threads)

    def set_verbose(self, verbose):
        deref(self.p).setVerbose(verbose)

    def solve(self):
        solution = Solution(self.num_variables)
        cdef string message
        deref(self.p).solve(solution.p[0], message)
        return solution, message

cdef class QuadraticSolver:

    cdef shared_ptr[decl.QuadraticSolverBackend] p
    cdef unsigned int num_variables

    def __cinit__(
            self,
            num_variables,
            default_variable_type,
            dict variable_types=None,
            Preference preference=Preference.Any):
        cdef decl.SolverFactory factory
        cdef cppmap[unsigned int, decl.VariableType] vtypes
        if variable_types is not None:
            for k, v in variable_types.items():
                vtypes[k] = v
        self.p = factory.createQuadraticSolverBackend(preference)
        self.num_variables = num_variables
        deref(self.p).initialize(num_variables, default_variable_type, vtypes)

    def set_objective(self, QuadraticObjective objective):
        deref(self.p).setObjective(objective.p[0])

    def set_constraints(self, LinearConstraints constraints):
        deref(self.p).setConstraints(constraints.p[0])

    def add_constraint(self, LinearConstraint constraint):
        deref(self.p).addConstraint(constraint.p[0])

    def set_timeout(self, timeout):
        deref(self.p).setTimeout(timeout)

    def set_optimality_gap(self, gap, absolute=False):
        deref(self.p).setOptimalityGap(gap, absolute)

    def set_num_threads(self, num_threads):
        deref(self.p).setNumThreads(num_threads)

    def solve(self):
        solution = Solution(self.num_variables)
        cdef string message
        deref(self.p).solve(solution.p[0], message)
        return solution, message



from libcpp cimport bool
from libcpp.string cimport string
from libcpp.memory cimport shared_ptr
from libcpp.map cimport map
from libcpp.vector cimport vector

cdef extern from 'impl/solvers/Relation.h':
    cdef enum Relation:
        LessEqual
        Equal
        GreaterEqual

cdef extern from 'impl/solvers/VariableType.h':
    cdef enum VariableType:
        Continuous
        Integer
        Binary

cdef extern from 'impl/solvers/Sense.h':
    cdef enum Sense:
        Minimize
        Maximize

cdef extern from 'impl/solvers/BackendPreference.h':
    cdef enum Preference:
        Any
        Scip
        Gurobi
        Cplex

cdef extern from 'impl/solvers/Solution.cpp':
    pass

cdef extern from 'impl/solvers/Solution.h':
    cdef cppclass Solution:
        Solution(unsigned int) except +
        void resize(unsigned int)
        unsigned int size()
        double& operator[](unsigned int i)
        void setValue(double value)
        double getValue()

cdef extern from 'impl/solvers/QuadraticObjective.cpp':
    pass

cdef extern from 'impl/solvers/QuadraticObjective.h':
    cdef cppclass QuadraticObjective:
        QuadraticObjective() except +
        QuadraticObjective(unsigned int) except +
        void setConstant(double)
        double getConstant()
        void setCoefficient(unsigned int, double)
        const vector[double]& getCoefficients()
        void setQuadraticCoefficient(unsigned int, unsigned int, double)
        void setSense(Sense)
        Sense getSense()
        void resize(unsigned int)
        unsigned int size()

cdef extern from 'impl/solvers/LinearObjective.h':
    cdef cppclass LinearObjective:
        LinearObjective() except +
        LinearObjective(unsigned int) except +
        void setConstant(double)
        double getConstant()
        void setCoefficient(unsigned int, double)
        const vector[double]& getCoefficients()
        void setSense(Sense)
        Sense getSense()
        void resize(unsigned int)
        unsigned int size()

cdef extern from 'impl/solvers/LinearConstraint.cpp':
    pass

cdef extern from 'impl/solvers/LinearConstraint.h':
    cdef cppclass LinearConstraint:
        LinearConstraint() except +
        void setCoefficient(unsigned int, double)
        const map[unsigned int, double]& getCoefficients()
        void setRelation(Relation)
        void setValue(double)
        Relation getRelation()
        double getValue()
        bool isViolated(const Solution&)

cdef extern from 'impl/solvers/LinearConstraints.cpp':
    pass

cdef extern from 'impl/solvers/LinearConstraints.h':
    cdef cppclass LinearConstraints:
        LinearConstraints() except +
        void clear()
        void add(LinearConstraint&)
        void addAll(LinearConstraints&);
        unsigned int size()

cdef extern from 'impl/solvers/LinearSolverBackend.h':
    cdef cppclass LinearSolverBackend:
        void initialize(unsigned int, VariableType, map[unsigned int, VariableType]&) except +
        void setObjective(LinearObjective&)
        void setConstraints(LinearConstraints&)
        void addConstraint(LinearConstraint&)
        void setTimeout(double)
        void setOptimalityGap(double, bool)
        void setNumThreads(unsigned int)
        void setVerbose(bool)
        bool solve(Solution& solution, string& message)

cdef extern from 'impl/solvers/QuadraticSolverBackend.h':
    cdef cppclass QuadraticSolverBackend:
        void initialize(unsigned int, VariableType, map[unsigned int, VariableType]&) except +
        void setObjective(QuadraticObjective&)
        void setConstraints(LinearConstraints&)
        void addConstraint(LinearConstraint&)
        void setTimeout(double)
        void setOptimalityGap(double, bool)
        void setNumThreads(unsigned int)
        bool solve(Solution& solution, string& message)

cdef extern from 'impl/solvers/ScipBackend.cpp':
    pass

cdef extern from 'impl/solvers/GurobiBackend.cpp':
    pass

cdef extern from 'impl/solvers/SolverFactory.cpp':
    pass

cdef extern from 'impl/solvers/SolverFactory.h':
    cdef cppclass SolverFactory:
        shared_ptr[LinearSolverBackend] createLinearSolverBackend(Preference) except +
        shared_ptr[QuadraticSolverBackend] createQuadraticSolverBackend(Preference) except +



from enum import IntEnum, auto

class Preference(IntEnum):
    Any = auto()
    Scip = auto()
    Gurobi = auto()
    Cplex = auto()

class VariableType(IntEnum):
    Continuous = auto()
    Integer = auto()
    Binary = auto()

class Sense(IntEnum):
    Minimize = auto()
    Maximize = auto()

class Relation(IntEnum):
    LessEqual = auto()
    Equal = auto()
    GreaterEqual = auto()

# cython appears to make these available as module-level constants as well
Any = Preference.Any
Scip = Preference.Scip
Gurobi = Preference.Gurobi
Cplex = Preference.Cplex
Continuous = VariableType.Continuous
Integer = VariableType.Integer
Binary = VariableType.Binary
Minimize = Sense.Minimize
Maximize = Sense.Maximize
Equal = Relation.Equal
GreaterEqual = Relation.GreaterEqual
LessEqual = Relation.LessEqual


class Solution:
    def __init__(self, size: int) -> None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, i: int) -> float: ...
    def __setitem__(self, i: int, value: float) -> None: ...
    def resize(self, size: int) -> None: ...
    def get_value(self) -> float: ...
    def set_value(self, value: float) -> None: ...

class LinearConstraint:
    def set_coefficient(self, i, value) -> None: ...
    def get_coefficients(self) -> None: ...
    def set_relation(self, relation) -> None: ...
    def set_value(self, value) -> None: ...
    def get_relation(self) -> None: ...
    def get_value(self) -> None: ...
    def is_violated(self, solution: Solution) -> None: ...

class LinearConstraints:
    def clear(self) -> None: ...
    def add(self, constraint: LinearConstraint) -> None: ...
    def add_all(self, constraints: LinearConstraints) -> None: ...
    def __len__(self) -> int: ...

class _Objective:
    def __init__(self, size: int = 0) -> None: ...
    def __len__(self) -> int: ...
    def resize(self, size: int) -> None: ...
    def set_sense(self, sense: Sense) -> None: ...
    def get_sense(self) -> Sense: ...
    def set_coefficient(self, i: int, value: float) -> None: ...
    def get_coefficients(self) -> list[float]: ...
    def get_constant(self) -> float: ...
    def set_constant(self, value: float) -> None: ...

class LinearObjective(_Objective): ...

class QuadraticObjective(_Objective):
    def set_quadratic_coefficient(self, i: int, j: int, value: float) -> None: ...

class _Solver:
    def __init__(
        self,
        num_variables: int,
        default_variable_type: VariableType,
        variable_types: dict[int, VariableType] | None = None,
        preference: Preference = Preference.Any,
    ) -> None: ...
    def set_constraints(self, constraints: LinearConstraints) -> None: ...
    def add_constraint(self, constraint: LinearConstraint) -> None: ...
    def set_timeout(self, timeout: float) -> None: ...
    def set_optimality_gap(self, gap: float, absolute: bool = False) -> None: ...
    def set_num_threads(self, num_threads: int) -> None: ...
    def solve(self) -> tuple[Solution, str]: ...

class LinearSolver(_Solver):
    def set_objective(self, objective: LinearObjective): ...
    def set_verbose(self, verbose: bool): ...

class QuadraticSolver(_Solver):
    def set_objective(self, objective: QuadraticObjective) -> None: ...



import os

from ctypes import util

from Cython.Build import cythonize
from setuptools import setup
from setuptools.extension import Extension

libraries = ["libscip"] if os.name == "nt" else ["scip"]
include_dirs = ['ilpy/impl']
library_dirs = []
compile_args = ["-O3", "-std=c++11", "-DHAVE_SCIP"]

# include conda environment windows include/lib if it exists
# this will be done automatically by conda build, but is useful if someone
# tries to build this directly with pip install in a conda environment
if os.name == "nt" and "CONDA_PREFIX" in os.environ:
    include_dirs.append(os.path.join(os.environ["CONDA_PREFIX"], "Library", "include"))
    library_dirs.append(os.path.join(os.environ["CONDA_PREFIX"], "Library", "lib"))

# look for various gurobi versions, which are annoyingly
# suffixed with the version number, and wildcards don't work
for v in ["100"]:
    GUROBI_LIB = f"libgurobi{v}" if os.name == 'nt' else f"gurobi{v}"
    if (gurolib := util.find_library(GUROBI_LIB)) is not None:
        print("FOUND GUROBI library: ", gurolib)
        libraries.append(GUROBI_LIB)
        compile_args.append("-DHAVE_GUROBI")
        break
else:
    print("WARNING: GUROBI library not found")


setup(
        name='ilpy',
        version='0.2.2',
        description='Python wrappers for popular MIP solvers.',
        url='https://github.com/funkelab/ilpy',
        author='Jan Funke',
        author_email='funkej@janelia.hhmi.org',
        license='MIT',
        packages=[
            'ilpy'
        ],
        ext_modules=cythonize([
            Extension(
                'ilpy.wrapper',
                sources=[
                    'ilpy/wrapper.pyx',
                ],
                extra_compile_args=compile_args,
                include_dirs=include_dirs,
                libraries=libraries,
                library_dirs=library_dirs,
                language='c++')
        ]),
        extras_require={
          "dev": ["flake8", "pytest", "pytest-cov"],
        },
        package_data={
            "ilpy": ["py.typed", "*.pyi"]
        },
)
